from .providers import LLMProviderFactory, GradeDocuments
from .constants import DEFAULT_MODEL


class DocumentGrader:
    def __init__(
        self, provider: str, api_key: str, model: str = DEFAULT_MODEL
    ):  # noqa: E501
        self.provider = LLMProviderFactory.create_provider(
            provider, api_key, model
        )  # noqa: E501

    def grade_document(self, document: str, question: str) -> GradeDocuments:
        return self.provider.grade_document(document, question)


def document_grader(
    provider: str, api_key: str, model: str = DEFAULT_MODEL
) -> DocumentGrader:
    return DocumentGrader(provider, api_key, model)



from .grader import document_grader, DocumentGrader
from .retriever import create_retriever
from .providers import LLMProviderFactory

__all__ = [
    "document_grader",
    "DocumentGrader",
    "create_retriever",
    "LLMProviderFactory",
]  # noqa: E501



from abc import ABC, abstractmethod
from langchain_core.pydantic_v1 import BaseModel, Field, SecretStr
from langchain_core.prompts import ChatPromptTemplate
from langchain_openai import ChatOpenAI
from langchain_anthropic import ChatAnthropic
from langchain_google_genai import ChatGoogleGenerativeAI
from .constants import (
    SYSTEM_PROMPT,
    DEFAULT_MODEL,
    REASONING_DESCRIPTION,
    BINARY_SCORE_DESCRIPTION,
)


class GradeDocuments(BaseModel):
    reasoning: str = Field(description=REASONING_DESCRIPTION)
    binary_score: str = Field(description=BINARY_SCORE_DESCRIPTION)


class LLMProvider(ABC):
    def __init__(self, api_key: str, model: str):
        self.llm = self._create_llm(api_key, model)
        self.prompt = self._create_prompt()

    @abstractmethod
    def _create_llm(self, api_key: str, model: str):
        pass

    def _create_prompt(self):
        return ChatPromptTemplate.from_messages(
            [
                ("system", SYSTEM_PROMPT),
                (
                    "human",
                    "Retrieved document: \n\n {document} \n\n User question: \n\n {question}",  # noqa: E501
                ),
            ]
        )

    def grade_document(self, document: str, question: str) -> GradeDocuments:
        structured_llm = self.llm.with_structured_output(GradeDocuments)
        chain = self.prompt | structured_llm
        return chain.invoke({"document": document, "question": question})


class OpenAIProvider(LLMProvider):
    def _create_llm(self, api_key: str, model: str):
        return ChatOpenAI(api_key=api_key, model=model)


class AnthropicProvider(LLMProvider):
    def _create_llm(self, api_key: str, model: str):
        return ChatAnthropic(api_key=api_key, model_name=model)


class GoogleProvider(LLMProvider):
    def _create_llm(self, api_key: str, model: str):
        return ChatGoogleGenerativeAI(
            google_api_key=SecretStr(api_key), model=model
        )  # noqa: E501


class LLMProviderFactory:
    @staticmethod
    def create_provider(
        provider: str, api_key: str, model: str = DEFAULT_MODEL
    ) -> LLMProvider:
        if provider == "openai":
            return OpenAIProvider(api_key, model)
        elif provider == "anthropic":
            return AnthropicProvider(api_key, model)
        elif provider == "google":
            return GoogleProvider(api_key, model)
        else:
            raise ValueError(f"Unsupported provider: {provider}")



from langchain.text_splitter import RecursiveCharacterTextSplitter
from langchain_community.document_loaders import WebBaseLoader
from langchain_community.vectorstores import Chroma
from langchain_openai import OpenAIEmbeddings


def create_retriever(urls: list[str], api_key: str):
    docs = [WebBaseLoader(url).load() for url in urls]
    docs_list = [item for sublist in docs for item in sublist]

    text_splitter = RecursiveCharacterTextSplitter.from_tiktoken_encoder(
        chunk_size=250, chunk_overlap=0
    )
    doc_splits = text_splitter.split_documents(docs_list)

    vectorstore = Chroma.from_documents(
        documents=doc_splits,
        collection_name="rag-chroma",
        embedding=OpenAIEmbeddings(api_key=api_key),
    )
    return vectorstore.as_retriever()



# flake8: noqa
SYSTEM_PROMPT = """
You are a grader assessing the relevance of a retrieved document to a user question.
Your goal is to filter out erroneous retrievals without being overly strict.
If the document contains keywords or semantic meanings related to the user question, grade it as relevant.
Give a binary score ('yes' or 'no') to indicate whether the document is relevant to the question.
"""

DEFAULT_MODEL = "gpt-3.5-turbo-0125"

REASONING_DESCRIPTION = "Thinking process to give a correct binary score shortly."
BINARY_SCORE_DESCRIPTION = "Documents are relevant to the question, 'yes' or 'no'"
